# Autogenerated by Astropy's setup.py on 2018-02-06 00:39:14
from __future__ import unicode_literals
import datetime

version = "2.0.4"
githash = "6cb6edc89b80fe417f65286167dc2db61a885ebc"


major = 2
minor = 0
bugfix = 4

release = True
timestamp = datetime.datetime(2018, 2, 6, 0, 39, 14)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
